#include "common.h"
#include "Label.h"
#include "LabelSettings.h"
#include "Font.h"
#include "Texture.h"
#include "SystemInfo.h"
#include "contextmenu.h"

#include <shellapi.h>
#include <shlobj.h>

#define TIMER_MOUSETRACK 1
#define TIMER_UPDATE 2
#define TIMER_SCROLL 3
#define TIMER_AUTOHIDE 4
#define TIMER_MOUSETRACK_MOVE 5

extern LabelList labelList;
extern LabelList desktoplabelList;
extern LabelList shortcutlabelList;

Label::Label(const string &name):height(0),width(0),x(0),y(0)
{
	this->name = name;

	mousePressed = false;
	mouseInside = false;

	movemodifierkeyPressed = false;
	moveButtonPressed = 0;

	hWnd = 0;
	hInstance = 0;
	box = 0;

	canBeVisible = true;
	visible = false;
	bUsingDefSkin = false;
	bInDestructor = false;

	autoWidthMode = 0;
	autoHeightMode = 0;

	scroll = false;
	scrollLimit = 0;
	scrollPosition = 0;
	
	backgroundDC = 0;
	bufferDC = 0;
	backgroundBitmap = 0;
	bufferBitmap = 0;

	hoverActive = false;
	pressedActive = false;

	useHover = false;
	usePressed = false;

	background = 0;
	font = 0;

	originalText.push_back("");
	current = 0;

	textChange = true;

	desktopItem = false;
	lineBreak = false;

	newlineCounter = 1;
}

Label::~Label()
{
	bInDestructor = true;

	if (hWnd && TooltipHints)
	{
		removeHint(hWnd);
		DestroyWindow(TooltipHints);
	}

	if(IsWindow(hWnd))
	{
		KillTimer(hWnd, TIMER_MOUSETRACK_MOVE);
		KillTimer(hWnd, TIMER_MOUSETRACK);
		KillTimer(hWnd, TIMER_UPDATE);
		KillTimer(hWnd, TIMER_SCROLL);

		DestroyWindow(hWnd);
	}

	if(backgroundBitmap)
	{
		backgroundBitmap = (HBITMAP) SelectObject(backgroundDC, backgroundBitmap);
		DeleteObject(backgroundBitmap);
	}
	if (backgroundDC)
		DeleteDC(backgroundDC);

	if(bufferBitmap)
	{
		bufferBitmap = (HBITMAP) SelectObject(bufferDC, bufferBitmap);
		DeleteObject(bufferBitmap);
	}
	if (bufferDC)
		DeleteDC(bufferDC);

	if(!bUsingDefSkin)
		delete background;

	if(font != defaultSettings->font)
		delete font;
}

void Label::load(HINSTANCE hInstance, HWND box)
{
	this->hInstance = hInstance;
	
	if (IsWindow(box))
	{
		this->box = box;
	}
	else
	{
		this->box = NULL;
	}
	
	if (IsWindow(hWnd))	// window already exists...
		return;

	hWnd = CreateWindowEx(box ? 0 : WS_EX_TOOLWINDOW,
		"xLabel",
		name.c_str(),
		box ? WS_CHILD : WS_POPUP,
		x, y,			// FIXME they're all 0, the window will be repositioned
		width, height,	// FIXME see above
		box ? box : GetLitestepWnd(), // GetLitestepDesktop(), FIXME: this way there aren't as many z-order problems =)
		0,
		hInstance,
		this);
		
	if (hWnd)
	{
		SetWindowLong(hWnd, GWL_USERDATA, MAGIC_DWORD);

		//ToolTips for Labels
		if ( !(GetRCString(name.c_str(), "Tooltip", "")).empty() )
		{
			TooltipHints = CreateWindow(TOOLTIPS_CLASS, NULL, TTS_ALWAYSTIP,
										  CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
										  NULL, NULL, hInstance, NULL);
			if (TooltipHints)
				SetWindowPos(TooltipHints, HWND_TOPMOST, 0, 0, 0, 0,
							 SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
		}

		reconfigure();
		repaint();
	}
	else
		delete this;
}

void Label::desktopload(HINSTANCE hInstance, string configLine)
{
	this->hInstance = hInstance;
	
    this->box = NULL;
	
	if (IsWindow(hWnd))	// window already exists...
		return;
		
	hWnd = CreateWindowEx(box ? 0 : WS_EX_TOOLWINDOW,
		"xLabel",
		name.c_str(),
		box ? WS_CHILD : WS_POPUP,
		x, y,			// FIXME they're all 0, the window will be repositioned
		width, height,	// FIXME see above
		box ? box : GetLitestepWnd(), // GetLitestepDesktop(), FIXME: this way there aren't as many z-order problems =)
		0,
		hInstance,
		this);
		
	if (hWnd)
	{
		SetWindowLong(hWnd, GWL_USERDATA, MAGIC_DWORD);
		
		//ToolTips for Icons
		TooltipHints = CreateWindow(TOOLTIPS_CLASS, NULL, TTS_ALWAYSTIP,
									  CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
									  NULL, NULL, hInstance, NULL);
		if (TooltipHints)
			SetWindowPos(TooltipHints, HWND_TOPMOST, 0, 0, 0, 0,
						 SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);

		desktopreconfigure(configLine);
		repaint(true);
	}
	else
		delete this;

}

void Label::shortcutload(HINSTANCE hInstance, string configLine, int lastx, int lasty)
{
	this->hInstance = hInstance;
	
    this->box = NULL;
	
	if (IsWindow(hWnd))	// window already exists...
		return;
		
	hWnd = CreateWindowEx(box ? 0 : WS_EX_TOOLWINDOW,
		"xLabel",
		name.c_str(),
		box ? WS_CHILD : WS_POPUP,
		x, y,			// FIXME they're all 0, the window will be repositioned
		width, height,	// FIXME see above
		box ? box : GetLitestepWnd(), // GetLitestepDesktop(), FIXME: this way there aren't as many z-order problems =)
		0,
		hInstance,
		this);
		
	if (hWnd)
	{
		SetWindowLong(hWnd, GWL_USERDATA, MAGIC_DWORD);

		//ToolTips for Shortcut2
		TooltipHints = CreateWindow(TOOLTIPS_CLASS, NULL, TTS_ALWAYSTIP,
									  CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
									  NULL, NULL, hInstance, NULL);
		if (TooltipHints)
			SetWindowPos(TooltipHints, HWND_TOPMOST, 0, 0, 0, 0,
						 SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);

		shortcutreconfigure(configLine, lastx, lasty);
		repaint();
	}
	else
		delete this;

}

void Label::reconfigure()
{
	LabelSettings settings(name.c_str());
	
	trueTransparency = settings.trueTransparency;
	alphaTransparency = settings.alphaTransparency;

	setBackground(settings.skin);		// FIXME: this does not check if old == new on !refresh
	setFont(settings.font);				// FIXME: see above
	
	reposition(settings.x, settings.y, settings.width, settings.height, 0, 0);	// FIXME? doesn't check old == new; ".0,0" at the end is from animation setup ;)

	tooltip = settings.tooltip;
	char tooltiptext[128];
	strcpy(tooltiptext, (systemInfo->processLabelText(tooltip, this, &dynamicText)).c_str());
	if (!tooltip.empty())
		addHint(this->hWnd, tooltiptext);
	strcpy(oldtooltiptext, tooltiptext);

	labelGroup = settings.labelGroup;

	bUseFahrenheit = settings.bUseFahrenheit;
	
	if ( !(GetRCString(name.c_str(), "HoverImage", "")).empty() )
		useHover = true;
	if ( !(GetRCString(name.c_str(), "PressedImage", "")).empty() )
		usePressed = true;

	setAlign(settings.align);
	setVertAlign(settings.vertAlign);
	setText(settings.text);

	alwaysUpdateContent = settings.alwaysUpdateContent;
	setUpdateInterval(settings.updateInterval);

	setLeftBorder(settings.leftBorder);
	setTopBorder(settings.topBorder);
	setRightBorder(settings.rightBorder);
	setBottomBorder(settings.bottomBorder);

	leftClickCommand = settings.leftClickCommand;
	leftDoubleClickCommand = settings.leftDoubleClickCommand;
	middleClickCommand = settings.middleClickCommand;
	middleDoubleClickCommand = settings.middleDoubleClickCommand;
	rightClickCommand = settings.rightClickCommand;
	rightDoubleClickCommand = settings.rightDoubleClickCommand;
	wheelDownCommand = settings.wheelDownCommand;
	wheelUpCommand = settings.wheelUpCommand;
	enterCommand = settings.enterCommand;
	leaveCommand = settings.leaveCommand;
	dropCommand = settings.dropCommand;
	textChangeCommand = settings.textChangeCommand;

	scrollPadLength = settings.scrollPadLength;
	scrollInterval = settings.scrollInterval;
	scrollSpeed = settings.scrollSpeed;
	setScrolling(settings.scroll);

	autoWidthMode = settings.autoWidthMode;
	autoHeightMode = settings.autoHeightMode;
	autoMinWidth = settings.width;
	autoMinHeight = settings.height;
	autoMaxWidth = settings.autoMaxWidth;
	autoMaxHeight = settings.autoMaxHeight;
	startCenterX = settings.x+(settings.width/2);
	startCenterY = settings.y+(settings.height/2);

	shadowY = font->shadowY;

	moveable = settings.moveable;
	moveKey = settings.moveKey;
	moveButton = settings.moveButton;
	
	labelLeftClickRegions = settings.labelLeftClickRegions;
	labelRightClickRegions = settings.labelRightClickRegions;
	labelMiddleClickRegions = settings.labelMiddleClickRegions;

	enterleaveRegion = settings.enterleaveRegion;

	if (hWnd)
	{
		UpdateWindowTransparent(hWnd, alphaTransparency);
		setGhosted(settings.ghosted);

		if (dropCommand.empty())
			DragAcceptFiles(hWnd, FALSE);
		else
			DragAcceptFiles(hWnd, TRUE);

		settings.startHidden ? hide() : show();
	}

	setAlwaysOnTop(settings.alwaysOnTop);	
}

void Label::desktopreconfigure(string configLine)
{
	NameValuePair iconTextValues[] = {
	{ "top", 1 },
	{ "bottom", 2 },
	{ "left", 3 },
	{ "right", 4 },
	{ 0, 0 }
	};

	ghosted = false;
	desktopItem = true;
	labelGroup = "xlabeldesktop";
	lineBreak = true;
	int iconSize = GetRCInt("xlabeldesktop", "IconSize", 32);
	int iconTextAlign = GetRCNamedValue("xlabeldesktop", "TextPosition", iconTextValues, 2);
	int iconTextSpacing = GetRCInt("xlabeldesktop" , "TextSpacing", 5);
	
	rememberPosition = GetRCBoolean("xlabeldesktop", "RememberPosition", false);
	configSavePath = GetRCString("xlabeldesktop", "ConfigFile", "");

	bUseFahrenheit = false;
	
	char name[32], ix[16], iy[16], path[128];
	char *buffers[] = {name, ix, iy};

	if(LCTokenize(configLine.c_str(), buffers, 3, path) >= 3)
	{
		this->width = GetRCInt("xlabeldesktop", "ItemWidth", 125);
		this->height = GetRCInt("xlabeldesktop", "ItemHeight", 50);
	}

	trueTransparency = true;
	alphaTransparency = GetRCInt("xlabeldesktop", "AlphaTransparency", 255, 0 , 255);

	this->x = ParseCoord(ix, 0, GetSystemMetrics(SM_CXSCREEN));
	this->y = ParseCoord(iy, 0, GetSystemMetrics(SM_CXSCREEN));

	setBackground( GetDesktopTexture(path, this->getName()) );		// FIXME: this does not check if old == new on !refresh
	setFont( GetRCFont("xlabeldesktop") );							// FIXME: see above

	reposition(this->x, this->y , this->width, this->height, 0, 0);

	string result;
	
	//Find Filename
	string s = path;
	int i = s.rfind("\\");
	i = (i < 0) ? 0 : i + 1;
	s = s.substr(i, s.length() - i);

	//Remove Extention
	i = s.find(".");
	i = (i < 0) ? s.length() : i;
	s = s.substr(0, i);

	result = s;

	setAlign(DT_CENTER);

	setText(result);

	autoWidthMode = 0;
	autoHeightMode = 0;

	alwaysUpdateContent = false;
	setUpdateInterval(1000);

	if (iconTextAlign == 1 || iconTextAlign == 2)
	{
		setLeftBorder(0);
		setRightBorder(0);
	}
	else if (iconTextAlign == 3)
	{
		setLeftBorder(0);
		setRightBorder(iconSize+iconTextSpacing+1);
	}
	else if (iconTextAlign == 4)
	{
		setLeftBorder(iconSize+iconTextSpacing+1);
		setRightBorder(0);
	}

	if (iconTextAlign == 1)
	{
		setTopBorder(0);
		setBottomBorder(iconSize+iconTextSpacing+1);
		setVertAlign(ALIGN_BOTTOM);
	}
	else if (iconTextAlign == 2)
	{
		setTopBorder(iconSize+iconTextSpacing+1);
		setBottomBorder(0);
		setVertAlign(ALIGN_TOP);
	}
	else if (iconTextAlign == 3 || iconTextAlign == 4)
	{
		setTopBorder(0);
		setBottomBorder(0);
		setVertAlign(ALIGN_CENTER);
	}

	iconCommand = path;
	//Strip '"' or anything else surrounding ;)
	iconCommand = iconCommand.substr(1, iconCommand.length() - 2);

	//ToolTip
	strcpy(path, "File: ");
	strcat(path, iconCommand.c_str());
	addHint(this->hWnd, path);

	leftClickCommand = "";
	leftDoubleClickCommand = ("!execute [\"" + iconCommand + "\"]");
	middleClickCommand = "!labelupdate xlabeldesktop";
	middleDoubleClickCommand = "";
	rightClickCommand = ".contextmenu";
	rightDoubleClickCommand = "";
	wheelDownCommand = "";
	wheelUpCommand = "";
	enterCommand = "";
	leaveCommand = "";
	dropCommand = "";
	textChangeCommand = "";

	scrollPadLength = 10;
	scrollInterval = 100;
	scrollSpeed = 1;
	setScrolling(false);

	shadowY = font->shadowY;

	moveable = true;
	moveKey = 1;
	moveButton = 1;

	//labelLeftClickRegions = settings.labelLeftClickRegions;
	//labelRightClickRegions = settings.labelRightClickRegions;
	//labelMiddleClickRegions = settings.labelMiddleClickRegions;

	enterleaveRegion = "0 0 -0 -0";

	if (hWnd)
	{
		UpdateWindowTransparent(hWnd, alphaTransparency);

		DragAcceptFiles(hWnd, false);
		GetRCBoolean("xlabeldesktop", "StartHidden") ? hide() : show();
	}

	setAlwaysOnTop(false);	

	HDC tmpDC = CreateCompatibleDC(NULL);
	HBITMAP tmpBmp = CreateBitmap(this->width, this->height, 1, 32, NULL);
	HGDIOBJ tmpObj = SelectObject(tmpDC, tmpBmp);

	background->apply(tmpDC, 0, 0, this->width, this->height, 0);

	SelectObject(tmpDC, tmpObj);

	// region handling a la MickeM =)
	HRGN region = BitmapToRegion(tmpBmp, RGB(254, 0, 254), 0, 0, 0);
	HRGN windowRgn = CreateRectRgn(0, 0, 0, 0);
	CombineRgn(windowRgn, region, NULL, RGN_COPY);

	if (!SetWindowRgn(hWnd, windowRgn, TRUE))
	{
		::DeleteObject(windowRgn);
	}

	DeleteObject(tmpBmp);
	DeleteObject(tmpObj);
	DeleteDC(tmpDC);		
}

void Label::shortcutreconfigure(string configLine, int lastx, int lasty)
{
	char name[32], ix[16], iy[16], normal[64], hover[64], pressed[64], rest[256];
	char *buffers[] = {name, ix, iy, normal, hover, pressed};

	LCTokenize(configLine.c_str(), buffers, 6, rest);

	if ( stricmp(hover, ".none") != 0 )
		useHover = true;
	if ( stricmp(pressed, ".none") != 0 )
		usePressed = true;
	
	//Initializing
	LPSTR fLCommand = "";
    LPSTR fRCommand = "";
    LPSTR fMCommand = "";
	LPSTR fDropCommand = "";

	bool IsVisible;
	bool AcceptInput;
	bool UseTooltip = GetRCBoolean("shortcut", "ShowCaptions", true);

	labelGroup = "0";
	char token[4096];
	LPCSTR nextToken = rest;
	bool hasCommand = true;

	if (GetToken(nextToken, token, &nextToken, false))
	{
		if (token[0] == '#')
		{
			//Lowercase
			strlwr(token);
			int tmpGroup = atoi(&token[1]);
			char buffer[16];
			labelGroup = _itoa( tmpGroup, buffer, 10 );
			AcceptInput = !strchr(token, 'k');
            IsVisible = !strchr(token, 'h');
			alwaysOnTop = strchr(token, 't') != 0;
			trueTransparency = !strchr(token, 'o') != 0;
			moveable = strchr(token, 'm') != 0;
			rememberPosition = strchr(token, 's') != 0;

			if (UseTooltip)
				UseTooltip = !strchr(token, 'c');
			else
				UseTooltip = strchr(token, 'c') != 0;
          
            hasCommand = GetToken(nextToken, token, &nextToken, false) != FALSE;
		}

		//Start Comaptibility Modus to filter out sound definitions
		if (StrIPos(token, ".wav"))
			hasCommand = GetToken(nextToken, token, &nextToken, false) != FALSE;
		else if (StrIPos(token, ".none"))
			hasCommand = GetToken(nextToken, token, &nextToken, false) != FALSE;
		else if (StrIPos(token, ".def"))
			hasCommand = GetToken(nextToken, token, &nextToken, false) != FALSE;
	
		if (StrIPos(token, ".wav"))
			hasCommand = GetToken(nextToken, token, &nextToken, false) != FALSE;
		else if (StrIPos(token, ".none"))
			hasCommand = GetToken(nextToken, token, &nextToken, false) != FALSE;
		else if (StrIPos(token, ".def"))
			hasCommand = GetToken(nextToken, token, &nextToken, false) != FALSE;
		//End Compatibility Modus to filter out sound definitions

        // ugly hack
        if (hasCommand)
        {
            char szCommands[4096];
            char szLCommand[4096] = { 0 };
            char szRCommand[4096] = { 0 };
            char szMCommand[4096] = { 0 };
            char szDropCommand[4096] = { 0 };
            
            if (nextToken)
            {
                sprintf(szCommands, "[%s] %s", token, nextToken);
            }
            else
            {
                sprintf(szCommands, "[%s]", token);
            }

            if (token[0] == '[')
            {
                LPSTR lpszCommands[] = { szLCommand, szRCommand, szMCommand, szDropCommand };
                CommandTokenize(szCommands, lpszCommands, 4, NULL);
            }
            else
            {
                // compatibility mode, leftclick only
                strcpy(szLCommand, szCommands);
            }
           
            if (szLCommand[0])
            {
                fLCommand = new char[strlen(szLCommand) + 1];
                strcpy(fLCommand, szLCommand);
            }

            if (szRCommand[0])
            {
                fRCommand = new char[strlen(szRCommand) + 1];
                strcpy(fRCommand, szRCommand);
            }

            if (szMCommand[0])
            {
                fMCommand = new char[strlen(szMCommand) + 1];
                strcpy(fMCommand, szMCommand);
            }

            if (szDropCommand[0])
            {
                fDropCommand = new char[strlen(szDropCommand) + 1];
                strcpy(fDropCommand, szDropCommand);
            }
            else if (szLCommand[0])
            {
                // default drag&drop command to leftclick command
                fDropCommand = new char[strlen(szLCommand) + 1];
                strcpy(fDropCommand, szLCommand);
            }
        }
	}

	configSavePath = GetRCString("shortcut", "ConfigFile", "");	

	string images = normal;
	images.append(" ");
	images.append(hover);
	images.append(" ");
	images.append(pressed);

	string trans;
	if (trueTransparency)
		trans = "true";
	else 
		trans = "false";

	setBackground( GetShortcutTexture(images, trans) );		// FIXME: this does not check if old == new on !refresh
	setFont( GetRCFont("") );								// FIXME: see above
	
	HBITMAP	checkbitmap = LoadLSImage(normal, 0);
	GetLSBitmapSize(checkbitmap, &width, &height);

	reposition(ParseCoord(ix, 0, GetSystemMetrics(SM_CXSCREEN), lastx), ParseCoord(iy, 0, GetSystemMetrics(SM_CYSCREEN), lasty), width, height, 0, 0);	// FIXME? doesn't check old == new; ".0,0" at the end is from animation setup ;)

	if (UseTooltip)
		addHint(this->hWnd, name);

	//bUseFahrenheit = settings.bUseFahrenheit;
	
	//setAlign(settings.align);
	//setVertAlign(settings.vertAlign);
	//setText(settings.text);

	//alwaysUpdateContent = settings.alwaysUpdateContent;
	//setUpdateInterval(settings.updateInterval);

	//setLeftBorder(settings.leftBorder);
	//setTopBorder(settings.topBorder);
	//setRightBorder(settings.rightBorder);
	//setBottomBorder(settings.bottomBorder);

	if (AcceptInput)
	{
		leftClickCommand = fLCommand;
		//leftDoubleClickCommand = settings.leftDoubleClickCommand;
		middleClickCommand = fMCommand;
		//middleDoubleClickCommand = settings.middleDoubleClickCommand;
		rightClickCommand = fRCommand;
		//rightDoubleClickCommand = settings.rightDoubleClickCommand;
		//wheelDownCommand = settings.wheelDownCommand;
		//wheelUpCommand = settings.wheelUpCommand;
		//enterCommand = settings.enterCommand;
		//leaveCommand = settings.leaveCommand;
		//dropCommand = fDropCommand;
		//textChangeCommand = settings.textChangeCommand;
	}
	else 
		setGhosted(true);

	//scrollPadLength = settings.scrollPadLength;
	//scrollInterval = settings.scrollInterval;
	//scrollSpeed = settings.scrollSpeed;
	//setScrolling(false);

	//shadowY = font->shadowY;

	moveKey = 1;
	moveButton = 1;

	//labelLeftClickRegions = settings.labelLeftClickRegions;
	//labelRightClickRegions = settings.labelRightClickRegions;
	//labelMiddleClickRegions = settings.labelMiddleClickRegions;

	enterleaveRegion = "0 0 -0 -0";

	if (hWnd)
	{
		UpdateWindowTransparent(hWnd, 255);

		DragAcceptFiles(hWnd, FALSE);
		IsVisible ? show() : hide();
	}

	setAlwaysOnTop(alwaysOnTop);

	//CleanUp
	if (fLCommand)
		delete[] fLCommand;
    if (fRCommand)
        delete[] fRCommand;
    if (fMCommand)
        delete[] fMCommand;
    if (fDropCommand)
        delete[] fDropCommand;
}

void Label::mzscriptvarcopy(string var)
{
	string bang = ("!varset " + var + " \"" + text + "\"");
	LSExecute(hWnd, bang.c_str(), SW_SHOWNORMAL);
}

void Label::posmzscriptvarcopy(string varx, string vary)
{
	if (box)
		return;

	string bang;
	string posx, posy;

	RECT cr;
	GetWindowRect(this->hWnd, &cr);

	char buffer[32];
	sprintf(buffer, "%d", cr.left);
	posx = buffer;
	sprintf(buffer, "%d", cr.top);
	posy = buffer;

	bang = ("!varset " + varx + " \"" + posx + "\"");
	LSExecute(hWnd, bang.c_str(), SW_SHOWNORMAL);

	bang = ("!varset " + vary + " \"" + posy + "\"");
	LSExecute(hWnd, bang.c_str(), SW_SHOWNORMAL);
}

void Label::sizemzscriptvarcopy(string varcx, string varcy)
{
	if (box)
		return;

	string bang;
	string cx, cy;

	RECT cr;
	GetWindowRect(this->hWnd, &cr);

	char buffer[32];
	sprintf(buffer, "%d", cr.right-cr.left);
	cx = buffer;
	sprintf(buffer, "%d", cr.bottom-cr.top);
	cy = buffer;

	bang = ("!varset " + varcx + " \"" + cx + "\"");
	LSExecute(hWnd, bang.c_str(), SW_SHOWNORMAL);

	bang = ("!varset " + varcy + " \"" + cy + "\"");
	LSExecute(hWnd, bang.c_str(), SW_SHOWNORMAL);
}

void Label::setAlwaysOnTop(bool alwaysOnTop)
{
	this->alwaysOnTop = alwaysOnTop;
	
	if (box)
		return;
	
	if(hWnd)
	{
		ModifyStyle(hWnd, WS_POPUP, WS_CHILD);
		SetParent(hWnd, alwaysOnTop ? 0 : GetLitestepDesktop());
		ModifyStyle(hWnd, WS_CHILD, WS_POPUP);
		
		SetWindowPos(hWnd, alwaysOnTop ? HWND_TOPMOST : HWND_NOTOPMOST,
			0, 0, 0, 0,
			SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE);
	}
}

void Label::setBox(HWND newparent)
{
	if (this->box == newparent)
		return;

	if (this->alwaysOnTop) this->setAlwaysOnTop(FALSE);
	
	this->box = newparent;
	
	ModifyStyle(hWnd, WS_POPUP, WS_CHILD);
	SetParent(hWnd, newparent);
//	ModifyStyle(hWnd, WS_CHILD, WS_POPUP);
}

void Label::setScrolling(bool scrolling)
{
	if (scrolling && !this->scroll)
	{
		scrollPosition = 0;
		SetTimer(hWnd, TIMER_SCROLL, scrollInterval, 0);
	}
	else if (!scrolling && this->scroll)
	{
		KillTimer(hWnd, TIMER_SCROLL);
		scrollPosition=0;
		repaint();
	}

	scrollLimit = 0;
	this->scroll = scrolling;
}

void Label::setScrollLimit(int limit)
{
	if (limit >= 0)
	{
		setScrolling(true);
		scrollLimit = limit + 1;
	}
	else
		setScrolling(false);

}

void Label::setBackground(Texture *background)
{
	if(!bUsingDefSkin)
		delete this->background;

	this->background = background;

	bUsingDefSkin = (background == defaultSettings->skin);
	
	repaint(true);
}

void Label::setFont(Font *font)
{
	if(this->font != defaultSettings->font)
		delete this->font;

	this->font = font;
	repaint();
}

void Label::setUpdateInterval(int updateInterval)
{
	this->updateInterval = updateInterval;
	if(hWnd && (dynamicText || alwaysUpdateContent))
		SetTimer(hWnd, TIMER_UPDATE, updateInterval, 0);
}

void Label::setAlign(int align)
{
	if (this->align == align)
		return;

	this->align = align;
	repaint();
}

void Label::setVertAlign(int vertAlign)
{
	if (this->vertAlign == vertAlign)
		return;

	this->vertAlign = vertAlign;
	repaint();
}

void Label::setText(const string &text)
{
	originalText = split(text, ";");
		
	current = 0;
	update();
}

void Label::setLeftBorder(int leftBorder)
{
	if (this->leftBorder == leftBorder)
		return;

	this->leftBorder = leftBorder;
	repaint();
}

void Label::setTopBorder(int topBorder)
{
	if (this->topBorder == topBorder)
		return;

	this->topBorder = topBorder;
	repaint();
}

void Label::setRightBorder(int rightBorder)
{
	if (this->rightBorder == rightBorder)
		return;

	this->rightBorder = rightBorder;
	repaint();
}

void Label::setBottomBorder(int bottomBorder)
{
	if (this->bottomBorder == bottomBorder)
		return;

	this->bottomBorder = bottomBorder;
	repaint();
}

void Label::setAlpha(int factor)
{
	if (factor >= 0 && factor <= 255)
		UpdateWindowTransparent(hWnd, factor);
}

void Label::setGhosted(bool ghosted)
{
	this->ghosted = ghosted;
	
	if (box)
		return;
	
	if(hWnd)
	{
		if (ghosted)
			ModifyStyleEx(hWnd, 0, WS_EX_TRANSPARENT);
		else
			ModifyStyleEx(hWnd, WS_EX_TRANSPARENT, 0);
	}
}

void Label::repaint(bool invalidateCache)
{
	if(hWnd)
	{
		if(invalidateCache)
		{
			if(backgroundDC && backgroundBitmap)
			{
				backgroundBitmap = (HBITMAP) SelectObject(backgroundDC, backgroundBitmap);
				DeleteObject(backgroundBitmap);
				backgroundBitmap = 0;
			}
		}

		InvalidateRect(hWnd, 0, FALSE);
	}
}

void Label::move(int x, int y, int steps, int time)
{
	if (steps > 0 && time > 0)
	{
	//---------------------------------
	int counter = 0;
	int deltax = x-this->x;
	int deltay = y-this->y;

	bool growx;
	bool growy;
	if (deltax < 0)
		growx = false;
	else
		growx = true;

	if (deltay < 0)
		growy = false;
	else
		growy = true;

	div_t div_result = div( deltax, steps );
    deltax = div_result.quot;
	div_result = div( deltay, steps );
    deltay = div_result.quot;

	if (deltax == 0)
		deltax = 1;
	if (deltay == 0)
		deltay = 1;

	int dynx = this->x;
	int dyny = this->y;

	this->x = x;
	this->y = y;

	if(hWnd)
	{
		while (counter <= steps)
		{
			dynx = dynx+deltax;
			if (growx && dynx > x)
				dynx = x;
			else if (!growx && dynx < x)
				dynx = x;

			dyny = dyny+deltay;
			if (growy && dyny > y)
				dyny = y;
			else if (!growy && dyny < y)
				dyny = y;
			
			SetWindowPos(hWnd, 0, dynx, dyny, 0, 0,
				SWP_NOSIZE | SWP_NOACTIVATE | SWP_NOZORDER);

			counter++;

			MSG msg;
			unsigned long lTime = GetTickCount();

			if( time > 0 )
			{
				while( (GetTickCount() - lTime) < time )
				{
					if( PeekMessage( &msg, NULL, 0, 0, PM_REMOVE ) )
					{
						TranslateMessage( &msg );
						DispatchMessage( &msg );
					}
				}
			}
		}
		SetWindowPos(hWnd, 0, x, y, 0, 0,
				SWP_NOSIZE | SWP_NOACTIVATE | SWP_NOZORDER);
	}
	//---------------------------------
	}
	else
	{
		this->x = x;
		this->y = y;

		if(hWnd)
		{
			SetWindowPos(hWnd, 0, x, y, 0, 0,
			SWP_NOSIZE | SWP_NOACTIVATE | SWP_NOZORDER);
		}
	}

	if (autoWidthMode != 0 || autoHeightMode != 0)
	{
		startCenterX = this->x+(this->width/2);
		startCenterY = this->y+(this->height/2);
	}
}

void Label::reposition(int x, int y, int width, int height, int steps, int time)
{
	if (steps > 0 && time > 0)
	{
	//---------------------------------
	int counter = 0;
	int deltax = x-this->x;
	int deltay = y-this->y;
	int deltasizex = width-this->width;
	int deltasizey = height-this->height;

	bool growx;
	bool growy;
	bool growsizex;
	bool growsizey;
	
	if (deltax < 0)
		growx = false;
	else
		growx = true;
	if (deltay < 0)
		growy = false;
	else
		growy = true;
	
	if (deltasizex < 0)
		growsizex = false;
	else
		growsizex = true;
	if (deltasizey < 0)
		growsizey = false;
	else
		growsizey = true;

	div_t div_result;
	div_result = div( deltax, steps );
    deltax = div_result.quot;
	div_result = div( deltay, steps );
    deltay = div_result.quot;

	div_result = div( deltasizex, steps );
    deltasizex = div_result.quot;
	div_result = div( deltasizey, steps );
    deltasizey = div_result.quot;

	if (deltax == 0)
		deltax = 1;
	if (deltay == 0)
		deltay = 1;
	if (deltasizex == 0)
		deltasizex = 1;
	if (deltasizey == 0)
		deltasizey = 1;

	int dynx = this->x;
	int dyny = this->y;
	int dynwidth = this->width;
	int dynheight = this->height;

	this->x = x;
	this->y = y;
	this->width = width;
	this->height = height;

	if(hWnd)
	{
		while (counter <= steps)
		{
			dynx = dynx+deltax;
			if (growx && dynx > x)
				dynx = x;
			else if (!growx && dynx < x)
				dynx = x;

			dyny = dyny+deltay;
			if (growy && dyny > y)
				dyny = y;
			else if (!growy && dyny < y)
				dyny = y;

			dynwidth = dynwidth+deltasizex;
			if (growsizex && dynwidth > width)
				dynwidth = width;
			else if (!growsizex && dynwidth < width)
				dynwidth = width;

			dynheight = dynheight+deltasizey;
			if (growsizey && dynheight > height)
				dynheight = height;
			else if (!growsizey && dynheight < height)
				dynheight = height;
			
			SetWindowPos(hWnd, 0, dynx, dyny, dynwidth, dynheight,
				SWP_NOACTIVATE | SWP_NOZORDER);

			counter++;

			MSG msg;
			unsigned long lTime = GetTickCount();

			if( time > 0 )
			{
				while( (GetTickCount() - lTime) < time )
				{
					if( PeekMessage( &msg, NULL, 0, 0, PM_REMOVE ) )
					{
						TranslateMessage( &msg );
						DispatchMessage( &msg );
					}
				}
			}
		}
		SetWindowPos(hWnd, 0, x, y, width, height,
			SWP_NOACTIVATE | SWP_NOZORDER);
	}
	//---------------------------------
	}
	else
	{
		this->x = x;
		this->y = y;
		this->height = height;
		this->width = width;

		if(hWnd)
		{
			SetWindowPos(hWnd, 0, x, y, width, height,
				SWP_NOACTIVATE | SWP_NOZORDER);
		}
	}

	if (trueTransparency && !desktopItem)
		updateRegion(0);

	if (autoWidthMode != 0 || autoHeightMode != 0)
	{
		startCenterX = this->x+(this->width/2);
		startCenterY = this->y+(this->height/2);
	}
}

void Label::resize(int width, int height, int steps, int time)
{
	if (steps > 0 && time > 0)
	{
	//---------------------------------
	int counter = 0;
	int deltax = width-this->width;
	int deltay = height-this->height;

	bool growx;
	bool growy;
	if (deltax < 0)
		growx = false;
	else
		growx = true;

	if (deltay < 0)
		growy = false;
	else
		growy = true;

	div_t div_result = div( deltax, steps );
    deltax = div_result.quot;
	div_result = div( deltay, steps );
    deltay = div_result.quot;

	if (deltax == 0)
		deltax = 1;
	if (deltay == 0)
		deltay = 1;

	int dynwidth = this->width;
	int dynheight = this->height;

	this->height = height;
	this->width = width;

	if(hWnd)
	{
		while (counter <= steps)
		{
			dynwidth = dynwidth+deltax;
			if (growx && dynwidth > width)
				dynwidth = width;
			else if (!growx && dynwidth < width)
				dynwidth = width;

			dynheight = dynheight+deltay;
			if (growy && dynheight > height)
				dynheight = height;
			else if (!growy && dynheight < height)
				dynheight = height;
			
			SetWindowPos(hWnd, 0, 0, 0, dynwidth, dynheight,
				SWP_NOMOVE | SWP_NOACTIVATE | SWP_NOZORDER);

			counter++;

			MSG msg;
			unsigned long lTime = GetTickCount();

			if( time > 0 )
			{
				while( (GetTickCount() - lTime) < time )
				{
					if( PeekMessage( &msg, NULL, 0, 0, PM_REMOVE ) )
					{
						TranslateMessage( &msg );
						DispatchMessage( &msg );
					}
				}
			}
		}
		SetWindowPos(hWnd, 0, 0, 0, width, height,
				SWP_NOMOVE | SWP_NOACTIVATE | SWP_NOZORDER);
	}
	//---------------------------------
	}
	else
	{
		this->height = height;
		this->width = width;

		if(hWnd)
		{
			SetWindowPos(hWnd, 0, 0, 0, width, height,
				SWP_NOMOVE | SWP_NOACTIVATE | SWP_NOZORDER);
		}
	}

	if (trueTransparency && !desktopItem)
		updateRegion(0);
}

void Label::internalHide()
{
	canBeVisible = false;

	if(hWnd)
	{
		ShowWindow(hWnd, SW_HIDE);
	}
}

void Label::internalShow()
{
	canBeVisible = true;

	if(hWnd && visible)
	{
		ShowWindow(hWnd, SW_SHOWNOACTIVATE);
	}
}

void Label::hide()
{
	if(hWnd)
	{
		visible = false;
		ShowWindow(hWnd, SW_HIDE);
		KillTimer(hWnd, TIMER_AUTOHIDE);
	}
}

void Label::show()
{
	if (!hWnd)
		load(hInstance, box);
	
	if(hWnd)
	{
		visible = true;
		KillTimer(hWnd, TIMER_AUTOHIDE);

		if(canBeVisible)
		{
			ShowWindow(hWnd, SW_SHOWNOACTIVATE);
		}
	}
}

void Label::showHide(double timeout)
{
	show();
	SetTimer(hWnd, TIMER_AUTOHIDE, timeout * 1000, NULL);
}

void Label::previous()
{
	current--;
	if(current < 0) current = originalText.size() - 1;
	update();
}

void Label::next()
{
	current++;
	if(current >= originalText.size()) current = 0;
	update();
}

void Label::update()
{
	string oldText = text;
	text = systemInfo->processLabelText(originalText[current], this, &dynamicText);
	
	//Start: Split in Multilines and check for longest line
	if(text != oldText)
	{
		vector<string> tmpLongest;
		int iter = 0;
		int longest = 0;
		string longestText;

		tmpLongest = split(text, "\n");
		longest = tmpLongest[iter].length();
		longestText = tmpLongest[iter];

		iter++;
		while (iter < tmpLongest.size())
		{
			if (longest < tmpLongest[iter].length())
			{
				longest = tmpLongest[iter].length();
				longestText = tmpLongest[iter];
			}
			iter++;
		}
		longestTextLine = longestText;
		newlineCounter = iter;

		textChange = true;
	}
	//End: Split in Multilines and check for longest line

	//Tooltips
	if (!tooltip.empty() && TooltipHints)
	{
		char tooltiptext[128];
		strcpy(tooltiptext, (systemInfo->processLabelText(tooltip, this, &dynamicText)).c_str());
		if ( stricmp(tooltiptext, oldtooltiptext) != 0 )
		{
			updateHint(this->hWnd, tooltiptext);
			strcpy(oldtooltiptext, tooltiptext);
		}
	}

	if(hWnd)
	{
		if(dynamicText || alwaysUpdateContent)
			SetTimer(hWnd, TIMER_UPDATE, updateInterval, 0);
		else
			KillTimer(hWnd, TIMER_UPDATE);
	}

	if(text != oldText)
	{
		if(!textChangeCommand.empty())
			LSExecute(NULL, textChangeCommand.c_str(), SW_SHOWNORMAL);
	}

	repaint();
}

void Label::clipboardCopy(string prefix)
{
	string toCopy = prefix.empty() ? text : (prefix + " " + text);
	SetClipboardText(hWnd, toCopy.c_str());
}

void Label::clipboardPaste(string prefix)
{
	char buffer[256];

	if(GetClipboardText(hWnd, buffer, 256))
	{
		string toPaste = prefix.empty() ? string(buffer) : (prefix + " " + string(buffer));
		setText(toPaste);
	}
}

void Label::relayMouseMessageToBox(UINT message, WPARAM wParam, LPARAM lParam)
{
	POINT pt;
	pt.x = GET_X_LPARAM(lParam);
	pt.y = GET_Y_LPARAM(lParam);

	if (box)
	{
		MapWindowPoints(hWnd, box, &pt, 1);
		PostMessage(box, message, wParam, MAKELPARAM((short) pt.x, (short) pt.y));
	}
	else
	{
		MapWindowPoints(hWnd, GetLitestepDesktop(), &pt, 1);
		PostMessage(GetLitestepDesktop(), message, wParam, MAKELPARAM((short) pt.x, (short) pt.y));
	}
}

void Label::onLButtonDblClk(int x, int y)
{
	if(leftDoubleClickCommand.length() > 0)
		LSExecute(hWnd, leftDoubleClickCommand.c_str(), SW_SHOWNORMAL);
	if (this->getBox() == 0)
		ReorderZOrder();
}

bool Label::checkRegion(int x, int y, StringList inputlist)
{
	bool found = false;
	if (!inputlist.empty())
	{
		for(StringListIterator it = inputlist.begin(); it != inputlist.end(); it++)
		{
			char left[16], top[16], right[16], bottom[16], command[128];
			char *buffers[] = {left, top, right, bottom};

			if(LCTokenize((*it).c_str(), buffers, 4, command) >= 4)
			{
				RECT rc;
				GetClientRect(hWnd, &rc);

				int intleft = ParseCoord(left, 0, rc.right-rc.left);
				int intright = ParseCoord(right, -0, rc.right-rc.left);
				int inttop = ParseCoord(top, 0, rc.bottom-rc.top);
				int intbottom  = ParseCoord(bottom, -0, rc.bottom-rc.top);

				if (intleft < intright && inttop < intbottom)
				{
					POINT pt;
					pt.x = x;
					pt.y = y;

					if(pt.x >= intleft && pt.x <= intright && pt.y >= inttop && pt.y <= intbottom)
					{
						if((*it).length() > 0)
						{	
							string temp = command;
							string bang = ("!execute [" + temp + "]");
							LSExecute(hWnd, bang.c_str(), SW_SHOWNORMAL);
							found = true;
							return found;
						}	
					}
				}
			}
		}
	}
	return found;
}

bool Label::checkenterleaveRegion(int x, int y)
{
	bool inside = false;
	char left[16], top[16], right[16], bottom[16];
	char *buffers[] = {left, top, right, bottom};
	if(LCTokenize(enterleaveRegion.c_str(), buffers, 4, NULL) >= 4)
	{
		POINT pt;
		pt.x = x;
		pt.y = y;

		RECT rc;
		GetClientRect(hWnd, &rc);

		int intleft = ParseCoord(left, 0, rc.right-rc.left);
		int intright = ParseCoord(right, -0, rc.right-rc.left);
		int inttop = ParseCoord(top, 0, rc.bottom-rc.top);
		int intbottom = ParseCoord(bottom, -0, rc.bottom-rc.top);

		if (intleft < intright && inttop < intbottom)
		{
			if(pt.x >= intleft && pt.x <= intright && pt.y >= inttop && pt.y <= intbottom)
				inside = true;
		}
	}

	return inside;
 }

void Label::onLButtonDown(int x, int y)
{
	if(moveable && moveButton == 1)
	{
		if(!mousePressed && !movemodifierkeyPressed)
			mousePressed = true;
		else if (!mousePressed && movemodifierkeyPressed)
		{
			if (box)
				return;
			mousePressed = true;
			moveButtonPressed = 1;
			GetCursorPos(&savedpt);
			ScreenToClient(hWnd, &savedpt);
			SetCapture(hWnd);
		}
	}
	else
	{
		if(!mousePressed)
			mousePressed = true;
	}

	if (usePressed)
	{
		hoverActive = false;
		pressedActive = true;
		repaint(true);
	
		if (trueTransparency && !desktopItem)
			updateRegion(2);
	}
	if (this->getBox() == 0)
		ReorderZOrder();
}

	
void Label::onLButtonUp(int x, int y)
{	
	if(moveable && moveButton == 1)
	{
		if(mousePressed && !movemodifierkeyPressed)
		{
			//To be on the save side
			ReleaseCapture();
			KillTimer(hWnd, TIMER_MOUSETRACK_MOVE);
			
			if (!checkRegion(x, y, labelLeftClickRegions))
			{
				POINT pt;
				pt.x = x;
				pt.y = y;

				RECT rc;
				GetClientRect(hWnd, &rc);
				this->x = rc.left;
				this->y = rc.top;

				if(PtInRect(&rc, pt))
				{
					if(leftClickCommand.length() > 0)
						LSExecute(hWnd, leftClickCommand.c_str(), SW_SHOWNORMAL);
				}
			}
			
			if ( rememberPosition )
			{
				string bang;
				bang = ("!labeldesktopsave " + this->name);
				LSExecute(hWnd, bang.c_str(), SW_SHOWNORMAL);
			}

			mousePressed = false;
			moveButtonPressed = 0;
		}
		else if (mousePressed && movemodifierkeyPressed)
		{
			//To be on the save side
			ReleaseCapture();
			KillTimer(hWnd, TIMER_MOUSETRACK_MOVE);

			if ( rememberPosition )
			{
				string bang;
				bang = ("!labeldesktopsave " + this->name);
				LSExecute(hWnd, bang.c_str(), SW_SHOWNORMAL);
			}

			mousePressed = false;
			moveButtonPressed = 0;
		}
	}
	else
	{
		//To be on the save side
		ReleaseCapture();
		KillTimer(hWnd, TIMER_MOUSETRACK_MOVE);

		if(mousePressed && !checkRegion(x, y, labelLeftClickRegions))
		{
			POINT pt;
			pt.x = x;
			pt.y = y;

			RECT rc;
			GetClientRect(hWnd, &rc);
			this->x = rc.left;
			this->y = rc.top;

			if(PtInRect(&rc, pt))
			{
				if(leftClickCommand.length() > 0)
					LSExecute(hWnd, leftClickCommand.c_str(), SW_SHOWNORMAL);
			}
		}
		mousePressed = false;
		moveButtonPressed = 0;
	}

	if (usePressed)
	{
		hoverActive = true;
		pressedActive = false;
		repaint(true);

		if (trueTransparency && !desktopItem)
			updateRegion(1);
	}
	if (this->getBox() == 0)	
		ReorderZOrder();
}

void Label::onMButtonDblClk(int x, int y)
{
	if(middleDoubleClickCommand.length() > 0)
		LSExecute(hWnd, middleDoubleClickCommand.c_str(), SW_SHOWNORMAL);
	if (this->getBox() == 0)
		ReorderZOrder();
}

void Label::onMButtonDown(int x, int y)
{
	if(moveable && moveButton == 3)
	{
		if(!mousePressed && !movemodifierkeyPressed)
			mousePressed = true;
		else if (!mousePressed && movemodifierkeyPressed)
		{
			if (box)
				return;
			mousePressed = true;
			moveButtonPressed = 3;
			GetCursorPos(&savedpt);
			ScreenToClient(hWnd, &savedpt);
			SetCapture(hWnd);
		}
	}
	else
	{
		if(!mousePressed)
			mousePressed = true;
	}

	if (usePressed)
	{
		hoverActive = false;
		pressedActive = true;
		repaint(true);
	
		if (trueTransparency && !desktopItem)
			updateRegion(2);
	}
	if (this->getBox() == 0)
		ReorderZOrder();
}

void Label::onMButtonUp(int x, int y)
{
	if(moveable && moveButton == 3)
	{
		if(mousePressed && !movemodifierkeyPressed)
		{
			//To be on the save side
			ReleaseCapture();
			KillTimer(hWnd, TIMER_MOUSETRACK_MOVE);
			
			if (!checkRegion(x, y, labelMiddleClickRegions))
			{
				POINT pt;
				pt.x = x;
				pt.y = y;

				RECT rc;
				GetClientRect(hWnd, &rc);
				this->x = rc.left;
				this->y = rc.top;

				if(PtInRect(&rc, pt))
				{
					if(middleClickCommand.length() > 0)
						LSExecute(hWnd, middleClickCommand.c_str(), SW_SHOWNORMAL);
				}
			}
			
			mousePressed = false;
			moveButtonPressed = 0;
		}
		else if (mousePressed && movemodifierkeyPressed)
		{
			//To be on the save side
			ReleaseCapture();
			KillTimer(hWnd, TIMER_MOUSETRACK_MOVE);

			mousePressed = false;
			moveButtonPressed = 0;
		}
	}
	else
	{
		//To be on the save side
		ReleaseCapture();
		KillTimer(hWnd, TIMER_MOUSETRACK_MOVE);

		if(mousePressed && !checkRegion(x, y, labelMiddleClickRegions))
		{
			POINT pt;
			pt.x = x;
			pt.y = y;

			RECT rc;
			GetClientRect(hWnd, &rc);
			this->x = rc.left;
			this->y = rc.top;

			if(PtInRect(&rc, pt))
			{
				if(middleClickCommand.length() > 0)
					LSExecute(hWnd, middleClickCommand.c_str(), SW_SHOWNORMAL);
			}
		}
		mousePressed = false;
		moveButtonPressed = 0;
	}

	if (usePressed)
	{
		hoverActive = true;
		pressedActive = false;
		repaint(true);

		if (trueTransparency && !desktopItem)
			updateRegion(1);
	}
	if (this->getBox() == 0)
		ReorderZOrder();
}

void Label::onRButtonDblClk(int x, int y)
{
	if(rightDoubleClickCommand.length() > 0)
		LSExecute(hWnd, rightDoubleClickCommand.c_str(), SW_SHOWNORMAL);
	if (this->getBox() == 0)
		ReorderZOrder();
}

void Label::onRButtonDown(int x, int y)
{
	if(moveable && moveButton == 2)
	{
		if(!mousePressed && !movemodifierkeyPressed)
			mousePressed = true;
		else if (!mousePressed && movemodifierkeyPressed)
		{
			if (box)
				return;
			mousePressed = true;
			moveButtonPressed = 2;
			GetCursorPos(&savedpt);
			ScreenToClient(hWnd, &savedpt);
			SetCapture(hWnd);
		}
	}
	else
	{
		if(!mousePressed)
			mousePressed = true;
	}

	if (usePressed)
	{
		hoverActive = false;
		pressedActive = true;
		repaint(true);
	
		if (trueTransparency && !desktopItem)
			updateRegion(2);
	}
	if (this->getBox() == 0)
		ReorderZOrder();
}

void Label::onRButtonUp(int x, int y)
{
	if(moveable && moveButton == 2)
	{
		if(mousePressed && !movemodifierkeyPressed)
		{
			//To be on the save side
			ReleaseCapture();
			KillTimer(hWnd, TIMER_MOUSETRACK_MOVE);
			
			if (!checkRegion(x, y, labelRightClickRegions))
			{
				POINT pt;
				pt.x = x;
				pt.y = y;

				RECT rc;
				GetClientRect(hWnd, &rc);
				this->x = rc.left;
				this->y = rc.top;

				if(PtInRect(&rc, pt))
				{
					if(rightClickCommand.length() > 0)
						LSExecute(hWnd, rightClickCommand.c_str(), SW_SHOWNORMAL);
				}
			}
			
			mousePressed = false;
			moveButtonPressed = 0;
		}
		else if (mousePressed && movemodifierkeyPressed)
		{
			//To be on the save side
			ReleaseCapture();
			KillTimer(hWnd, TIMER_MOUSETRACK_MOVE);

			mousePressed = false;
			moveButtonPressed = 0;
		}
	}
	else 
	{
		//To be on the save side
		ReleaseCapture();
		KillTimer(hWnd, TIMER_MOUSETRACK_MOVE);

		if(mousePressed && !checkRegion(x, y, labelRightClickRegions))
		{
			POINT pt;
			pt.x = x;
			pt.y = y;

			RECT rc;
			GetClientRect(hWnd, &rc);
			this->x = rc.left;
			this->y = rc.top;

			if(PtInRect(&rc, pt))
			{
				if(rightClickCommand.length() > 0 && rightClickCommand != ".contextmenu")
					LSExecute(hWnd, rightClickCommand.c_str(), SW_SHOWNORMAL);
				else
					ContextMenu(iconCommand.c_str());
			}
		}
		mousePressed = false;
		moveButtonPressed = 0;
	}

	if (usePressed)
	{
		hoverActive = true;
		pressedActive = false;
		repaint(true);

		if (trueTransparency && !desktopItem)
			updateRegion(1);
	}
	if (this->getBox() == 0)
		ReorderZOrder();
}

void Label::onWheelDown(int x, int y)
{
	if(wheelDownCommand.length() > 0)
		LSExecute(hWnd, wheelDownCommand.c_str(), SW_SHOWNORMAL);
	if (this->getBox() == 0)
		ReorderZOrder();
}

void Label::onWheelUp(int x, int y)
{
	if(wheelUpCommand.length() > 0)
		LSExecute(hWnd, wheelUpCommand.c_str(), SW_SHOWNORMAL);
	if (this->getBox() == 0)
		ReorderZOrder();
}

void Label::onMouseEnter()
{
	if(enterCommand.length() > 0)
		LSExecute(hWnd, enterCommand.c_str(), SW_SHOWNORMAL);

	if (useHover)
	{
		hoverActive = true;
		pressedActive = false;
		repaint(true);

		if (trueTransparency && !desktopItem)
			updateRegion(1);
	}
}

void Label::onMouseLeave()
{
	if(leaveCommand.length() > 0)
		LSExecute(hWnd, leaveCommand.c_str(), SW_SHOWNORMAL);

	if (useHover)
	{
		hoverActive = false;
		pressedActive = false;
		repaint(true);

		if (trueTransparency && !desktopItem)
			updateRegion(0);
	}
}

void Label::onMouseMove(int x, int y)
{
	if(moveable)
	{
		if (mousePressed && movemodifierkeyPressed && moveButtonPressed == moveButton)
			SetTimer(hWnd, TIMER_MOUSETRACK_MOVE, 2, 0);
		else
		{
			if (autoWidthMode != 0 || autoHeightMode != 0)
			{
				startCenterX = this->x+(this->width/2);
				startCenterY = this->y+(this->height/2);
			}

			KillTimer(hWnd, TIMER_MOUSETRACK_MOVE);
			KillTimer(hWnd, TIMER_MOUSETRACK);

			if(!mouseInside)
			{
				if(checkenterleaveRegion(x, y))
				{
					mouseInside = true;
					onMouseEnter();
				}
			}

			SetTimer(hWnd, TIMER_MOUSETRACK, 100, 0);
		}
	}
	else if (!moveable)
	{
		if(!mouseInside)
		{
			if(checkenterleaveRegion(x, y))
			{
				mouseInside = true;
				onMouseEnter();
			}
		}

		SetTimer(hWnd, TIMER_MOUSETRACK, 100, 0);
	}
}

void Label::onPaint(HDC hDC)
{
	RECT r;
	GetWindowRect(hWnd, &r);

	int width = r.right - r.left;
	int height = r.bottom - r.top;

	// keep a cached rendition of the background
	if(!backgroundBitmap)
	{
		if(!backgroundDC)
			backgroundDC = CreateCompatibleDC(hDC);

		backgroundBitmap = CreateCompatibleBitmap(hDC, width, height);
		backgroundBitmap = (HBITMAP) SelectObject(backgroundDC, backgroundBitmap);

		if(background->isTransparent())
		{
			if(box)
			{
				// save the previous DC contents as the background
				BitBlt(backgroundDC, 0, 0, width, height, hDC, 0, 0, SRCCOPY);
			}
			else
			{
				// paint desktop on display DC and then into background buffer
				PaintDesktopEx(backgroundDC, 0, 0, width, height, r.left, r.top, FALSE);
			}
		}
		
		if (hoverActive && useHover)
			background->apply(backgroundDC, 0, 0, width, height, 1);
		else if (pressedActive && usePressed)
			background->apply(backgroundDC, 0, 0, width, height, 2);
		else
			background->apply(backgroundDC, 0, 0, width, height, 0);

	}

	// double buffer for flicker-free paint
	if(!bufferBitmap)
	{
		if(!bufferDC)
			bufferDC = CreateCompatibleDC(hDC);

		bufferBitmap = CreateCompatibleBitmap(hDC, width, height);
		bufferBitmap = (HBITMAP) SelectObject(bufferDC, bufferBitmap);
	}

	// blt background into double buffer
	BitBlt(bufferDC, 0, 0, width, height, backgroundDC, 0, 0, SRCCOPY);

	// render text
	long textWidth;
	long textHeight;

	//font->measure(bufferDC, text, 0, &textWidth, &textHeight);
	font->measure(bufferDC, longestTextLine, 0, &textWidth, &textHeight);

	//don't break line if text fits into label
	if ( desktopItem && lineBreak && textWidth > width - leftBorder - rightBorder)
	{
		string s = text;
		string e;

		//before
		int i = s.rfind(" ");
		i = (i < 0) ? 0 : i + 1;
		e = s.substr(i, s.length() - i);
		//after
		i = s.rfind(" ");
		i = (i < 0) ? s.length() : i ;
		s = s.substr(0, i);

		//combine and set
		if (s != e)
			setText((s + "\\n" + e));

		lineBreak = false;
	}

	if ( (autoWidthMode != 0 || autoHeightMode != 0) && textChange )
	{ 
		int deltax;
		int deltay;

		if (autoWidthMode != 0)
			deltax = textWidth - (width - leftBorder - rightBorder);
		else
			deltax = 0;

		if (autoHeightMode != 0)
			deltay = (newlineCounter*textHeight) - (height - topBorder - bottomBorder);
		else
			deltay = 0;

		int newx = x;
		int newy = y;
		int newwidth = width + deltax;
		int newheight = height + deltay;

		//Check MIN/MAX
		if (newwidth < autoMinWidth || newwidth > autoMaxWidth)
		{
			if (newwidth < autoMinWidth)
			{
				deltax = deltax + (autoMinWidth-newwidth);
				newwidth = autoMinWidth;
			}
			if (newwidth > autoMaxWidth)
			{
				deltax = deltax - (newwidth-autoMaxWidth);
				newwidth = autoMaxWidth;
			}
		}
		if (newheight < autoMinHeight || newheight > autoMaxHeight)
		{
			if (newheight < autoMinHeight)
			{
				deltay = deltay + (autoMinHeight-newheight);
				newheight = autoMinHeight;
			}
			if (newheight > autoMaxHeight)
			{
				deltay = deltay - (newheight-autoMaxHeight);
				newheight = autoMaxHeight;
			}
		}

		if ((autoWidthMode == 0 || autoWidthMode == 1) && (autoHeightMode == 0 || autoHeightMode == 1))
		{
			newx = x;
			newy = y;
		}
		else if ((autoWidthMode == 0 || autoWidthMode == 1) && autoHeightMode == 2)
		{
			newx = x;
			newy = y-deltay;
		}
		else if ((autoWidthMode == 0 || autoWidthMode == 1) && autoHeightMode == 3)
		{
			newx = x;
			newy = startCenterY-(newheight/2);
		}
		else if (autoWidthMode == 2 && (autoHeightMode == 0 || autoHeightMode == 1))
		{
			newx = x-deltax;
			newy = y;
		}
		else if (autoWidthMode == 2 && autoHeightMode == 2)
		{
			newx = x-deltax;
			newy = y-deltay;
		}
		else if (autoWidthMode == 2 && autoHeightMode == 3)
		{
			newx = x-deltax;
			newy = startCenterY-(newheight/2);
		}
		else if (autoWidthMode == 3 && (autoHeightMode == 0 || autoHeightMode == 1))
		{
			newx = startCenterX-(newwidth/2);
			newy = y;
		}
		else if (autoWidthMode == 3 && autoHeightMode == 2)
		{
			newx = startCenterX-(newwidth/2);
			newy = y-deltay;
		}
		else if (autoWidthMode == 3 && autoHeightMode == 3)
		{
			newx = startCenterX-(newwidth/2);
			newy = startCenterY-(newheight/2);
		}

		reposition(newx, newy , newwidth, newheight, 0, 0);
		
		font->apply(bufferDC,
			leftBorder,
			topBorder,
			width - leftBorder - rightBorder,
			height - topBorder - bottomBorder,
			text,
			align /* | DT_SINGLELINE */ | DT_VCENTER | DT_NOPREFIX | DT_EXPANDTABS,
			vertAlign);

		textChange = false;
	}
	//don't scroll if text fits into label
	else if(!(scroll && textWidth > width - leftBorder - rightBorder))
	{
		font->apply(bufferDC,
			leftBorder,
			topBorder,
			width - leftBorder - rightBorder,
			height - topBorder - bottomBorder,
			text,
			align /* | DT_SINGLELINE */ | DT_VCENTER | DT_NOPREFIX | DT_EXPANDTABS,
			vertAlign);
	}
	else if (scroll)
	{
		if(scrollPosition <= 0)
		{
			scrollPosition = textWidth + scrollPadLength;
			
			if (scrollLimit)
			{
				scrollLimit--;
				if (scrollLimit == 0)
					setScrolling(false);
			}
		}

		if(scrollPosition > textWidth + scrollPadLength )
		{
			scrollPosition = 0;
		}

		HRGN oldRGN = NULL;
		HRGN newRGN = NULL;
		GetClipRgn(bufferDC, oldRGN);
		// positive shadows
		if (shadowY >= 0 && (shadowY-bottomBorder)<0 )
		{
			newRGN=CreateRectRgn(leftBorder, topBorder, width - rightBorder, height - bottomBorder + shadowY); // V
		}
		else if (shadowY >= 0)
		{
			newRGN=CreateRectRgn(leftBorder, topBorder, width - rightBorder, height - bottomBorder); // V
		}
		// handle negative shadows
		else if (shadowY<0 && topBorder+shadowY >= 0)
		{
			newRGN=CreateRectRgn(leftBorder, topBorder + shadowY, width - rightBorder, height - bottomBorder); // V
		}
		else
		{
			newRGN=CreateRectRgn(leftBorder, 0, width - rightBorder, height - bottomBorder); // V
		}
		
		SelectClipRgn(bufferDC, newRGN);
		DeleteObject(newRGN);

		font->apply(bufferDC,
			leftBorder, //(width - textWidth) / 2 - 2,
			topBorder,
			scrollPosition - scrollPadLength,
			height - topBorder - bottomBorder,
			text,
			DT_RIGHT | DT_VCENTER | DT_NOPREFIX | DT_EXPANDTABS,
			vertAlign);

		font->apply(bufferDC,
			scrollPosition + leftBorder,
			topBorder,
			textWidth,
			height - topBorder - bottomBorder,
			text,
			DT_LEFT | DT_VCENTER | DT_NOPREFIX | DT_EXPANDTABS,
			vertAlign);

		SelectClipRgn(bufferDC, oldRGN);
		DeleteObject(oldRGN);
	};

	// blt the double buffer to the display
	BitBlt(hDC, 0, 0, width, height, bufferDC, 0, 0, SRCCOPY);
}

void Label::onSize(int width, int height)
{
	this->height = height;
	this->width = width;

	if(backgroundDC && backgroundBitmap)
	{
		backgroundBitmap = (HBITMAP) SelectObject(backgroundDC, backgroundBitmap);
		DeleteObject(backgroundBitmap);
		backgroundBitmap = 0;
	}

	if(bufferDC && bufferBitmap)
	{
		bufferBitmap = (HBITMAP) SelectObject(bufferDC, bufferBitmap);
		DeleteObject(bufferBitmap);
		bufferBitmap = 0;
	}

	InvalidateRect(hWnd, 0, FALSE);
}

void Label::onTimer(int timerID)
{
	switch(timerID)
	{
		case TIMER_MOUSETRACK:

			POINT pt;
			GetCursorPos(&pt);
			ScreenToClient(hWnd, &pt);

			if(!checkenterleaveRegion(pt.x, pt.y))
			{
				if (mouseInside)
				{
					onMouseLeave();	
					mouseInside = false;
				}
			}

			RECT rc;
			GetWindowRect(hWnd, &rc);
			GetCursorPos(&pt);

			if(!PtInRect(&rc, pt))
			{
				KillTimer(hWnd, TIMER_MOUSETRACK);
				mouseInside = false;
				mousePressed = false;
			}

		break;

		case TIMER_MOUSETRACK_MOVE:

			POINT newpt;
			GetCursorPos(&newpt);
			
			newpt.x = newpt.x-savedpt.x;
			newpt.y = newpt.y-savedpt.y;
			move(newpt.x, newpt.y, 0, 0);
		
		break;

		case TIMER_UPDATE:

			update();

		break;

		case TIMER_SCROLL:

			scrollPosition -= scrollSpeed;
			repaint();

		break;

		case TIMER_AUTOHIDE:

			hide();

		break;
	}
}

void Label::onWindowPosChanged(WINDOWPOS *windowPos)
{
	repaint(true);
}

bool Label::onWindowMessage(UINT message, WPARAM wParam, LPARAM lParam, LRESULT &lResult)
{
	extern HWND messageHandler;
	
	if (!desktopItem)
	{
		if (moveKey == 3)
			movemodifierkeyPressed = (0x8000 & GetKeyState(VK_CONTROL)) && (0x8000 & GetKeyState(VK_SHIFT));
		else if (moveKey == 2)
			movemodifierkeyPressed = (0x8000 & GetKeyState(VK_SHIFT));
		else if (moveKey == 1)
			movemodifierkeyPressed = (0x8000 & GetKeyState(VK_CONTROL));
		else 
			movemodifierkeyPressed = (0x8000 & GetKeyState(VK_CONTROL));
	}
	else
		movemodifierkeyPressed = true;

	if (moveable && !movemodifierkeyPressed)
		KillTimer(hWnd, TIMER_MOUSETRACK_MOVE);

	switch(message)
	{
		case LM_SETLABELTEXT:
		{
			setText(string((const char *) lParam));
			return true;
		}

		case WM_COPYDATA:
		{
			COPYDATASTRUCT *cds = (COPYDATASTRUCT *) lParam;
			
			if(cds->dwData == LM_SETLABELTEXT)
				setText(string((const char *) cds->lpData));

			return false;
		}

		case WM_CLOSE:
		{
			lResult = 0;
			return true;
		}

		case WM_DESTROY:
		{
			if(box)		// FIXME: this is the old workaround
			{
				box = NULL;
				hide();
			}
				
			hWnd = NULL;	// FIXME: dirty hack
			
			return false;
		}

		case WM_DROPFILES:
		{
			if(dropCommand.empty())	// we shouldn't get here... but somehow we did :p
				break;

			char *szSourceFiles = NULL;
			char buffer[MAX_PATH], szTarget[MAX_PATH];
			int numFiles;
			UINT len,scursor = 0;
			int intFileAction;

			char action[32];
			char *buffers[] = {action};

			LCTokenize(dropCommand.c_str(), buffers, 1, szTarget);

			if (stricmp(action, "move") == 0)
				intFileAction = 1;
			else if (stricmp(action, "copy") == 0)
				intFileAction = 2;
			else if (stricmp(action, "delete") == 0)
				intFileAction = 0;

			string targetPath = szTarget;
			targetPath = targetPath.substr(1, targetPath.length() - 2);

			numFiles = DragQueryFile((HDROP) wParam, 0xFFFFFFFF, NULL, NULL);

			szSourceFiles = new char[MAX_PATH*numFiles+1];
			scursor = 0;

			for (int i=0;i<numFiles;i++)
			{
				len = DragQueryFile((HDROP) wParam, i, buffer, MAX_PATH);
				
				strcpy(szSourceFiles+scursor,buffer);
				scursor += len + 1;
			}
				
			DragFinish((HDROP) wParam);

			string check = (targetPath + "*");
			
			//Test Path for Existence
			WIN32_FIND_DATA FindData;
			HANDLE hFindHandle = FindFirstFile(check.c_str(), &FindData);

			if(hFindHandle != INVALID_HANDLE_VALUE)
			{
				szSourceFiles[scursor] = 0;
				strcpy(buffer, targetPath.c_str());
				buffer[strlen(buffer)+1] = 0;
				
				SHFILEOPSTRUCT shfop;
				shfop.hwnd = /*hWnd*/NULL;
				if (intFileAction == 1)
					shfop.wFunc = FO_MOVE;
				else if (intFileAction == 2)
					shfop.wFunc = FO_COPY;
				else if (intFileAction == 0)
					shfop.wFunc = FO_DELETE;
				shfop.pFrom = szSourceFiles;
				shfop.pTo = buffer;
				shfop.fFlags = FOF_NOCONFIRMMKDIR;
				SHFileOperation(&shfop);
			}
			
			delete[] szSourceFiles;
			
			return 0;
		}
		
		case WM_LBUTTONDBLCLK:
		{
			if(leftDoubleClickCommand.empty())
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onLButtonDblClk((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_LBUTTONDOWN:
		{
			//Andymon Moving Extention
			if(leftDoubleClickCommand.empty() && leftClickCommand.empty() && labelLeftClickRegions.empty() && !moveable)
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onLButtonDown((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_LBUTTONUP:
		{
			//Andymon Moving Extention
			if(leftDoubleClickCommand.empty() && leftClickCommand.empty() && labelLeftClickRegions.empty() && !moveable)
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onLButtonUp((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_MBUTTONDBLCLK:
		{
			if(middleDoubleClickCommand.empty())
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onMButtonDblClk((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_MBUTTONDOWN:
		{
			//Andymon Moving Extention
			if(middleDoubleClickCommand.empty() && middleClickCommand.empty() && labelMiddleClickRegions.empty() && !moveable)
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onMButtonDown((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_MBUTTONUP:
		{
			//Andymon Moving Extention
			if(middleDoubleClickCommand.empty() && middleClickCommand.empty() && labelMiddleClickRegions.empty() && !moveable)
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onMButtonUp((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_RBUTTONDBLCLK:
		{
			if(rightDoubleClickCommand.empty())
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onRButtonDblClk((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_RBUTTONDOWN:
		{
			//Andymon Moving Extention
			if(rightDoubleClickCommand.empty() && rightClickCommand.empty() && labelRightClickRegions.empty() && !moveable)
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onRButtonDown((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_RBUTTONUP:
		{
			//Andymon Moving Extention
			if(rightDoubleClickCommand.empty() && rightClickCommand.empty() && labelRightClickRegions.empty() && !moveable)
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onRButtonUp((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_MOUSEWHEEL:
		{
			if ((short)(HIWORD(wParam)) < 0)
			{
				if(wheelDownCommand.empty())
				{
					relayMouseMessageToBox(message, wParam, lParam);
					return true;
				}

				onWheelDown((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
				return true;
			}
			else
			{
				if(wheelUpCommand.empty())
				{
					relayMouseMessageToBox(message, wParam, lParam);
					return true;
				}
				
				onWheelUp((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
				return true;
			}
		}

		case WM_MOUSEMOVE:
		{
			onMouseMove((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_PAINT:
		{
			PAINTSTRUCT ps;
			HDC hDC;

			if(!wParam)
				hDC = BeginPaint(hWnd, &ps);
			else
				hDC = (HDC) wParam;

			onPaint(hDC);

			if(!wParam)
				EndPaint(hWnd, &ps);

			return true;
		}

		case WM_SIZE:
		{
			onSize((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_TIMER:
		{
			onTimer((int) wParam);
			return true;
		}

		case WM_WINDOWPOSCHANGED:
		{
			onWindowPosChanged((WINDOWPOS *) lParam);
			return false;
		}
	}

	return false;
}

LRESULT Label::windowProcedure(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	Label *label = NULL;

	if(message == WM_NCCREATE)
	{
		label = (Label *) ((CREATESTRUCT *) lParam)->lpCreateParams;
		label->hWnd = hWnd;
		SetWindowLong(hWnd, 0, (LONG) label);
	}
	else
		label = (Label *) GetWindowLong(hWnd, 0);

	if(label)
	{
		LRESULT lResult = 0;

		if(label->onWindowMessage(message, wParam, lParam, lResult))
			return lResult;
	}

	return DefWindowProc(hWnd, message, wParam, lParam);
}

//Tooltips
void Label::addHint(HWND hWnd, LPSTR caption)
{
	TOOLINFO ti;
	RECT clientRect;

	if (!TooltipHints)
		return ;
	GetClientRect(hWnd, &clientRect);
	ti.cbSize = sizeof(TOOLINFO);
	ti.uFlags = TTF_SUBCLASS;
	ti.hwnd = hWnd;
	ti.uId = 0;
	ti.rect = clientRect;
	ti.hinst = hInstance;
	ti.lpszText = caption;
	ti.lParam = 0;
	SendMessage(TooltipHints, TTM_ADDTOOL, 0, LPARAM(&ti));
}

void Label::updateHint(HWND hWnd, LPSTR caption)
{
	TOOLINFO ti;
	RECT clientRect;

	if (!TooltipHints)
		return ;
	GetClientRect(hWnd, &clientRect);
	ti.cbSize = sizeof(TOOLINFO);
	ti.uFlags = TTF_SUBCLASS;
	ti.hwnd = hWnd;
	ti.uId = 0;
	ti.rect = clientRect;
	ti.hinst = hInstance;
	ti.lpszText = caption;
	ti.lParam = 0;
	SendMessage(TooltipHints, TTM_UPDATETIPTEXT, 0, LPARAM(&ti));
}

void Label::removeHint(HWND hWnd)
{
	TOOLINFO ti;
	RECT emptyRect = {0, 0, 0, 0};

	if (!TooltipHints)
		return ;
	ti.cbSize = sizeof(TOOLINFO);
	ti.uFlags = 0;
	ti.hwnd = hWnd;
	ti.uId = 0;
	ti.rect = emptyRect;
	ti.hinst = hInstance;
	ti.lpszText = NULL;
	ti.lParam = 0;
	SendMessage(TooltipHints, TTM_DELTOOL, 0, LPARAM(&ti));
}

BOOL Label::UpdateWindowTransparent(HWND hWnd, unsigned char factor)
{
   /* First, see if we can get the API call we need. If we've tried
    * once, we don't need to try again. */
   if (!initialized)
   {
      HMODULE hDLL = LoadLibrary ("user32");

      pSetLayeredWindowAttributes = 
         (PSLWA) GetProcAddress(hDLL, "SetLayeredWindowAttributes");

      initialized = TRUE;
   }

   if (pSetLayeredWindowAttributes == NULL)
      return FALSE;
   else
	   ModifyStyleEx(hWnd, 0, WS_EX_LAYERED);

   /* Windows need to be layered to be made transparent. This is done
    * by modifying the extended style bits to contain WS_EX_LAYERED. */
   SetLastError(0);

   if (GetLastError())
      return FALSE;

   /* Now, we need to set the 'layered window attributes'. This
    * is where the alpha values get set. */
   
   return pSetLayeredWindowAttributes (hWnd, 
                                       RGB(255,0,255), 
                                       factor,
                                       LWA_COLORKEY|LWA_ALPHA);
}

void Label::ReorderZOrder()
{
	LabelListIterator iter;
	for( iter = labelList.begin(); iter != labelList.end(); iter++)
	{
		if((*iter)->name.c_str() == this->name.c_str())
		{
			iter++;
			if (iter != NULL)
			{
				SetWindowPos(this->hWnd, (*iter)->hWnd,
					0, 0, 0, 0,
					SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE);
			}
			break;
		}
	}
	for( iter = shortcutlabelList.begin(); iter != shortcutlabelList.end(); iter++)
	{
		if((*iter)->name.c_str() == this->name.c_str())
		{
			iter++;
			if (iter != NULL)
			{
				SetWindowPos(this->hWnd, (*iter)->hWnd,
					0, 0, 0, 0,
					SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE);
			}
			break;
		}
	}
}

void Label::updateRegion(int drawImage)
{
	HDC tmpDC = CreateCompatibleDC(NULL);
	HBITMAP tmpBmp = CreateBitmap(width, height, 1, 32, NULL);
	HGDIOBJ tmpObj = SelectObject(tmpDC, tmpBmp);

	//Draw Image: 0 normal, 1 hover, 2 pressed
	background->apply(tmpDC, 0, 0, width, height, drawImage);

	SelectObject(tmpDC, tmpObj);

	// region handling a la MickeM =)
	HRGN region = BitmapToRegion(tmpBmp, RGB(255, 0, 255), 0x101010, 0, 0);
	HRGN windowRgn = CreateRectRgn(0, 0, 0, 0);
	CombineRgn(windowRgn, region, NULL, RGN_COPY);

	if (!SetWindowRgn(hWnd, windowRgn, TRUE))
	{
		::DeleteObject(windowRgn);
	}

	DeleteObject(tmpBmp);
	DeleteObject(tmpObj);
	DeleteDC(tmpDC);		
}